import { defu } from 'defu';
import { resolvePath } from 'mlly';
import { resolveAlias, addPluginTemplate, addTemplate, addWebpackPlugin, useLogger, defineNuxtModule, isNuxt2 } from '@nuxt/kit';
import * as Sentry from '@sentry/node';
import { Integrations, withScope, captureException } from '@sentry/node';
import { fileURLToPath } from 'url';
import { resolve, posix } from 'path';
import { relative } from 'pathe';
import * as PluggableIntegrations from '@sentry/integrations';

const boolToText = (value) => value ? "enabled" : "disabled";
const envToBool = (env) => Boolean(env && env.toLowerCase() !== "false" && env !== "0");
const canInitialize = (options) => Boolean(options.initialize && options.dsn);
const clientSentryEnabled = (options) => !options.disabled && !options.disableClientSide;
const serverSentryEnabled = (options) => !options.disabled && !options.disableServerSide;
function callOnce(fn) {
  let called = false;
  return function callOnceWrapper(...subargs) {
    if (!called) {
      called = true;
      return fn(...subargs);
    }
  };
}

const BROWSER_INTEGRATIONS = ["Breadcrumbs", "Dedupe", "FunctionToString", "GlobalHandlers", "HttpContext", "InboundFilters", "LinkedErrors", "TryCatch"];
const BROWSER_PLUGGABLE_INTEGRATIONS = ["CaptureConsole", "Debug", "ExtraErrorData", "HttpClient", "ReportingObserver", "RewriteFrames"];
const SERVER_INTEGRATIONS = ["Console", "ContextLines", "FunctionToString", "Http", "InboundFilters", "LinkedErrors", "LocalVariables", "Modules", "OnUncaughtException", "OnUnhandledRejection", "RequestData"];
const SERVER_PLUGGABLE_INTEGRATIONS = ["CaptureConsole", "Debug", "Dedupe", "ExtraErrorData", "RewriteFrames", "Transaction"];
function filterDisabledIntegrations(integrations) {
  return getIntegrationsKeys(integrations).filter((key) => integrations[key]);
}
function getIntegrationsKeys(integrations) {
  return Object.keys(integrations);
}
function isBrowserDefaultIntegration(name) {
  return BROWSER_INTEGRATIONS.includes(name);
}
function isBrowserPluggableIntegration(name) {
  return BROWSER_PLUGGABLE_INTEGRATIONS.includes(name);
}
function isServerDefaultIntegration(name) {
  return SERVER_INTEGRATIONS.includes(name);
}
function isServerPlugabbleIntegration(name) {
  return SERVER_PLUGGABLE_INTEGRATIONS.includes(name);
}
async function getApiMethods(packageName) {
  const packageApi = await import(packageName);
  const apiMethods = [];
  for (const key in packageApi) {
    if (key === "default") {
      for (const subKey in packageApi[key]) {
        if (typeof packageApi[key][subKey] === "function") {
          apiMethods.push(subKey);
        }
      }
      continue;
    }
    if (typeof packageApi[key] === "function") {
      apiMethods.push(key);
    }
  }
  return apiMethods;
}
async function resolveRelease(moduleOptions) {
  if (!("release" in moduleOptions.config)) {
    try {
      const SentryCli = await import('@sentry/cli').then((m) => m.default || m);
      const cli = new SentryCli();
      return (await cli.releases.proposeVersion()).trim();
    } catch {
    }
  }
}
function resolveLazyOptions(options, apiMethods, logger) {
  if (options.lazy) {
    const defaultLazyOptions = {
      injectMock: true,
      injectLoadHook: false,
      mockApiMethods: true,
      chunkName: "sentry",
      webpackPrefetch: false,
      webpackPreload: false
    };
    options.lazy = defu(options.lazy, defaultLazyOptions);
    if (!options.lazy.injectMock) {
      options.lazy.mockApiMethods = [];
    } else if (options.lazy.mockApiMethods === true) {
      options.lazy.mockApiMethods = apiMethods;
    } else if (Array.isArray(options.lazy.mockApiMethods)) {
      const mockMethods = options.lazy.mockApiMethods;
      options.lazy.mockApiMethods = mockMethods.filter((method) => apiMethods.includes(method));
      const notfoundMethods = mockMethods.filter((method) => !apiMethods.includes(method));
      if (notfoundMethods.length) {
        logger.warn("Some specified methods to mock weren't found in @sentry/vue:", notfoundMethods);
      }
      if (!options.lazy.mockApiMethods.includes("captureException")) {
        options.lazy.mockApiMethods.push("captureException");
      }
    }
  }
}
function resolveTracingOptions(options, config) {
  if (!options.tracing) {
    return;
  }
  const defaultTracingOptions = {
    tracesSampleRate: 1,
    browserTracing: {},
    vueOptions: {
      trackComponents: true
    }
  };
  const userOptions = typeof options.tracing === "boolean" ? {} : options.tracing;
  const tracingOptions = defu(userOptions, defaultTracingOptions);
  if (config.tracesSampleRate === void 0) {
    config.tracesSampleRate = tracingOptions.tracesSampleRate;
  }
  options.tracing = tracingOptions;
  options.serverIntegrations = defu(options.serverIntegrations, { Http: { tracing: true } });
}
async function resolveClientOptions(nuxt, moduleOptions, logger) {
  const options = moduleOptions;
  let config = defu({}, options.config);
  let clientConfigPath;
  if (typeof options.clientConfig === "string") {
    clientConfigPath = resolveAlias(options.clientConfig);
    clientConfigPath = relative(nuxt.options.buildDir, clientConfigPath);
  } else {
    config = defu(options.clientConfig, options.config);
  }
  const apiMethods = await getApiMethods("@sentry/vue");
  resolveLazyOptions(options, apiMethods, logger);
  resolveTracingOptions(options, config);
  for (const name of getIntegrationsKeys(options.clientIntegrations)) {
    if (!isBrowserDefaultIntegration(name) && !isBrowserPluggableIntegration(name)) {
      logger.warn(`Sentry clientIntegration "${name}" is not recognized and will be ignored.`);
      delete options.clientIntegrations[name];
    }
  }
  let customClientIntegrations;
  if (options.customClientIntegrations) {
    if (typeof options.customClientIntegrations === "string") {
      customClientIntegrations = resolveAlias(options.customClientIntegrations);
      customClientIntegrations = relative(nuxt.options.buildDir, customClientIntegrations);
    } else {
      logger.warn(`Invalid customClientIntegrations option. Expected a file path, got "${typeof options.customClientIntegrations}".`);
    }
  }
  return {
    BROWSER_INTEGRATIONS,
    BROWSER_PLUGGABLE_INTEGRATIONS,
    dev: nuxt.options.dev,
    runtimeConfigKey: options.runtimeConfigKey,
    config: {
      dsn: options.dsn,
      ...config
    },
    clientConfigPath,
    lazy: options.lazy,
    apiMethods,
    customClientIntegrations,
    logMockCalls: options.logMockCalls,
    // for mocked only
    tracing: options.tracing,
    initialize: canInitialize(options),
    integrations: filterDisabledIntegrations(options.clientIntegrations).reduce((res, key) => {
      res[key] = options.clientIntegrations[key];
      return res;
    }, {})
  };
}
async function resolveServerOptions(nuxt, moduleOptions, logger) {
  const options = moduleOptions;
  for (const name of getIntegrationsKeys(options.serverIntegrations)) {
    if (!isServerDefaultIntegration(name) && !isServerPlugabbleIntegration(name)) {
      logger.warn(`Sentry serverIntegration "${name}" is not recognized and will be ignored.`);
      delete options.serverIntegrations[name];
    }
  }
  let customIntegrations = [];
  if (options.customServerIntegrations) {
    const resolvedPath = resolveAlias(options.customServerIntegrations);
    try {
      customIntegrations = (await import(resolvedPath).then((m) => m.default || m))();
      if (!Array.isArray(customIntegrations)) {
        logger.error(`Invalid value returned from customServerIntegrations plugin. Expected an array, got "${typeof customIntegrations}".`);
      }
    } catch (error) {
      logger.error(`Error handling the customServerIntegrations plugin:
${error}`);
    }
  }
  const defaultConfig = {
    dsn: options.dsn,
    integrations: [
      ...filterDisabledIntegrations(options.serverIntegrations).map((name) => {
        const opt = options.serverIntegrations[name];
        try {
          if (isServerDefaultIntegration(name)) {
            return Object.keys(opt).length ? new Integrations[name](opt) : new Integrations[name]();
          } else if (isServerPlugabbleIntegration(name)) {
            return Object.keys(opt).length ? new PluggableIntegrations[name](opt) : new PluggableIntegrations[name]();
          } else {
            throw new Error(`Unsupported server integration "${name}"`);
          }
        } catch (error) {
          throw new Error(`Failed initializing server integration "${name}".
${error}`);
        }
      }),
      ...customIntegrations
    ]
  };
  let serverConfig = options.serverConfig;
  if (typeof serverConfig === "string") {
    const resolvedPath = resolveAlias(serverConfig);
    try {
      serverConfig = (await import(resolvedPath).then((m) => m.default || m))();
    } catch (error) {
      logger.error(`Error handling the serverConfig plugin:
${error}`);
    }
  }
  const config = defu(defaultConfig, options.config, options.serverConfig, getRuntimeConfig(nuxt, options));
  const apiMethods = await getApiMethods("@sentry/node");
  resolveLazyOptions(options, apiMethods, logger);
  resolveTracingOptions(options, options.config);
  return {
    config,
    apiMethods,
    lazy: options.lazy,
    logMockCalls: options.logMockCalls,
    // for mocked only
    tracing: options.tracing
  };
}
function getRuntimeConfig(nuxt, options) {
  const { publicRuntimeConfig } = nuxt.options;
  const { runtimeConfigKey } = options;
  if (publicRuntimeConfig && typeof publicRuntimeConfig !== "function" && runtimeConfigKey in publicRuntimeConfig) {
    return defu(publicRuntimeConfig[runtimeConfigKey].config, publicRuntimeConfig[runtimeConfigKey].serverConfig);
  }
}

const RESOLVED_RELEASE_FILENAME = "sentry.release.config.mjs";
async function buildHook(nuxt, moduleOptions, logger) {
  const release = await resolveRelease(moduleOptions);
  const templateDir = fileURLToPath(new URL("./templates", import.meta.url));
  const pluginOptionClient = clientSentryEnabled(moduleOptions) ? moduleOptions.lazy ? "lazy" : "client" : "mocked";
  const clientOptions = defu({ config: { release } }, await resolveClientOptions(nuxt, moduleOptions, logger));
  addPluginTemplate({
    src: resolve(templateDir, `plugin.${pluginOptionClient}.js`),
    filename: "sentry.client.js",
    mode: "client",
    options: clientOptions
  });
  const pluginOptionServer = serverSentryEnabled(moduleOptions) ? "server" : "mocked";
  const serverOptions = defu({ config: { release } }, await resolveServerOptions(nuxt, moduleOptions, logger));
  addPluginTemplate({
    src: resolve(templateDir, `plugin.${pluginOptionServer}.js`),
    filename: "sentry.server.js",
    mode: "server",
    options: serverOptions
  });
  if (serverSentryEnabled(moduleOptions)) {
    addTemplate({
      src: resolve(templateDir, "options.ejs"),
      filename: RESOLVED_RELEASE_FILENAME,
      options: { release }
    });
  }
  if (!clientOptions.dev && !clientOptions.config.debug) {
    const webpack = await import('webpack').then((m) => m.default || m);
    addWebpackPlugin(new webpack.DefinePlugin({
      __SENTRY_DEBUG__: "false"
    }));
  }
}
async function webpackConfigHook(nuxt, webpackConfigs, options, logger) {
  let WebpackPlugin;
  try {
    WebpackPlugin = await import('@sentry/webpack-plugin').then((m) => m.default || m);
  } catch {
    throw new Error('The "@sentry/webpack-plugin" package must be installed as a dev dependency to use the "publishRelease" option.');
  }
  const publishRelease = defu({}, options.publishRelease);
  const nuxtOptions = nuxt.options;
  if (!publishRelease.urlPrefix) {
    if (typeof nuxtOptions.router.base === "string" && typeof nuxtOptions.build.publicPath === "string") {
      const publicPath = posix.join(nuxtOptions.router.base, nuxtOptions.build.publicPath);
      publishRelease.urlPrefix = publicPath.startsWith("/") ? `~${publicPath}` : publicPath;
    }
  }
  if (!Array.isArray(publishRelease.include)) {
    const { include } = publishRelease;
    publishRelease.include = [...include ? [include] : []];
  }
  const { buildDir } = nuxtOptions;
  if (!options.disableServerRelease) {
    publishRelease.include.push(`${buildDir}/dist/server`);
  }
  if (!options.disableClientRelease) {
    publishRelease.include.push(`${buildDir}/dist/client`);
  }
  publishRelease.release = options.config.release || publishRelease.release || await resolveRelease(options);
  if (!publishRelease.release) {
    logger.warn('Sentry release will not be published because "config.release" was not set nor it was possible to determine it automatically from the repository');
    return;
  }
  const attachCommits = envToBool(process.env.SENTRY_AUTO_ATTACH_COMMITS);
  if (attachCommits) {
    publishRelease.setCommits = publishRelease.setCommits || {};
    const { setCommits } = publishRelease;
    if (setCommits.auto === void 0) {
      setCommits.auto = true;
    }
    const repo = process.env.SENTRY_RELEASE_REPO || "";
    if (repo && setCommits.repo === void 0) {
      setCommits.repo = repo;
    }
  }
  for (const config2 of webpackConfigs) {
    config2.devtool = options.sourceMapStyle;
  }
  const config = webpackConfigs[webpackConfigs.length - 1];
  config.plugins = config.plugins || [];
  config.plugins.push(new WebpackPlugin(publishRelease));
}
async function initializeServerSentry(nuxt, moduleOptions, sentryHandlerProxy, logger) {
  if (process.sentry) {
    return;
  }
  let release;
  try {
    const path = resolve(nuxt.options.buildDir, RESOLVED_RELEASE_FILENAME);
    release = (await import(path)).release;
  } catch {
  }
  const serverOptions = await resolveServerOptions(nuxt, moduleOptions, logger);
  const config = defu({ release }, serverOptions.config);
  if (canInitialize(moduleOptions)) {
    Sentry.init(config);
    sentryHandlerProxy.errorHandler = Sentry.Handlers.errorHandler();
    sentryHandlerProxy.requestHandler = Sentry.Handlers.requestHandler(moduleOptions.requestHandlerConfig);
    if (serverOptions.tracing) {
      await import('@sentry/tracing');
      sentryHandlerProxy.tracingHandler = Sentry.Handlers.tracingHandler();
    }
  }
  process.sentry = Sentry;
}
async function shutdownServerSentry() {
  if (process.sentry) {
    await process.sentry.close();
    process.sentry = void 0;
  }
}

const logger = useLogger("nuxt:sentry");
const module = defineNuxtModule({
  meta: {
    name: "@nuxtjs/sentry",
    configKey: "sentry",
    compatibility: {
      nuxt: "^2.15.8"
    }
  },
  defaults: (nuxt) => ({
    lazy: false,
    dsn: process.env.SENTRY_DSN || "",
    disabled: envToBool(process.env.SENTRY_DISABLED) || false,
    initialize: envToBool(process.env.SENTRY_INITIALIZE) || true,
    runtimeConfigKey: "sentry",
    disableClientSide: envToBool(process.env.SENTRY_DISABLE_CLIENT_SIDE) || false,
    disableServerSide: envToBool(process.env.SENTRY_DISABLE_SERVER_SIDE) || false,
    publishRelease: envToBool(process.env.SENTRY_PUBLISH_RELEASE) || false,
    disableServerRelease: envToBool(process.env.SENTRY_DISABLE_SERVER_RELEASE) || false,
    disableClientRelease: envToBool(process.env.SENTRY_DISABLE_CLIENT_RELEASE) || false,
    logMockCalls: true,
    sourceMapStyle: "source-map",
    tracing: false,
    clientIntegrations: {
      ExtraErrorData: {},
      ReportingObserver: {},
      RewriteFrames: {}
    },
    serverIntegrations: {
      Dedupe: {},
      ExtraErrorData: {},
      RewriteFrames: {},
      Transaction: {}
    },
    customClientIntegrations: "",
    customServerIntegrations: "",
    config: {
      environment: nuxt.options.dev ? "development" : "production"
    },
    serverConfig: {},
    clientConfig: {},
    requestHandlerConfig: {}
  }),
  async setup(options, nuxt) {
    const defaultsPublishRelease = {
      include: [],
      ignore: [
        "node_modules",
        ".nuxt/dist/client/img"
      ],
      configFile: ".sentryclirc"
    };
    if (options.publishRelease) {
      options.publishRelease = defu(options.publishRelease, defaultsPublishRelease);
    }
    if (canInitialize(options) && (clientSentryEnabled(options) || serverSentryEnabled(options))) {
      const status = `(client side: ${boolToText(clientSentryEnabled(options))}, server side: ${boolToText(serverSentryEnabled(options))})`;
      logger.success(`Sentry reporting is enabled ${status}`);
    } else {
      let why;
      if (options.disabled) {
        why = '"disabled" option has been set';
      } else if (!options.dsn) {
        why = "no DSN has been provided";
      } else if (!options.initialize) {
        why = '"initialize" option has been set to false';
      } else {
        why = "both client and server side clients are disabled";
      }
      logger.info(`Sentry reporting is disabled (${why})`);
    }
    if (clientSentryEnabled(options)) {
      const clientDependencies = ["lodash.mergewith", "@sentry/integrations", "@sentry/vue", ...options.tracing ? ["@sentry/tracing"] : []];
      for (const dep of clientDependencies) {
        nuxt.options.alias[`~${dep}`] = (await resolvePath(dep)).replace(/\/cjs\//, "/esm/");
      }
    }
    if (serverSentryEnabled(options)) {
      const sentryHandlerProxy = {
        errorHandler: (error, _, __, next) => {
          next(error);
        },
        requestHandler: (_, __, next) => {
          next();
        },
        tracingHandler: (_, __, next) => {
          next();
        }
      };
      nuxt.hook("render:setupMiddleware", (app) => app.use((req, res, next) => {
        sentryHandlerProxy.requestHandler(req, res, next);
      }));
      if (options.tracing) {
        nuxt.hook("render:setupMiddleware", (app) => app.use((req, res, next) => {
          sentryHandlerProxy.tracingHandler(req, res, next);
        }));
      }
      nuxt.hook("render:errorMiddleware", (app) => app.use((error, req, res, next) => {
        sentryHandlerProxy.errorHandler(error, req, res, next);
      }));
      nuxt.hook("generate:routeFailed", ({ route, errors }) => {
        errors.forEach(({ error }) => withScope((scope) => {
          scope.setExtra("route", route);
          captureException(error);
        }));
      });
      if (isNuxt2()) {
        const initHook = nuxt.options._build ? "build:compile" : "ready";
        nuxt.hook(initHook, () => initializeServerSentry(nuxt, options, sentryHandlerProxy, logger));
        const shutdownServerSentryOnce = callOnce(() => shutdownServerSentry());
        nuxt.hook("generate:done", shutdownServerSentryOnce);
        nuxt.hook("close", shutdownServerSentryOnce);
      }
    }
    nuxt.hook("build:before", () => buildHook(nuxt, options, logger));
    if (options.publishRelease && !options.disabled && !nuxt.options.dev) {
      if (isNuxt2()) {
        nuxt.hook("webpack:config", (webpackConfigs) => webpackConfigHook(nuxt, webpackConfigs, options, logger));
      }
    }
  }
});

export { module as default };
